/*
    Profile
    copyright (c) 2002-2003 Kazuki IWAMOTO http://www.maid.org/ iwm@maid.org

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "fileio.h"
#include "misc.h"
#include "profile.h"


/******************************************************************************
*                                                                             *
* եؿ                                                        *
*                                                                             *
******************************************************************************/
/*	եꥹȲ
	file,ե̾
	 RET,ꥹ																*/
static ProfileList *
profile_list (const gchar *file)
{
	gchar buf[256], *data = NULL, *tmp, *section = NULL;
	gint n;
	FileIO *fio;
	ProfileList *p = NULL, *q, *r;

	fio = fileio_open (file, FILEIO_ACCESS_READ, FILEIO_SHARE_READ,
													FILEIO_MODE_OPEN_EXISTING);
	if (fio == NULL)
		return NULL;
	while (fileio_gets (buf, 256, fio) != NULL)
	  {
		if (data == NULL)
		  {
			data = g_strdup (buf);
		  }
		else
		  {
			tmp = data;
			data = g_strconcat (tmp, buf, NULL);
			g_free (tmp);
		  }
		n = data != NULL ? g_strlen (data) : 0;
		if (n > 0 && data[n - 1] == '\n')
		  {
			data[n - 1] = '\0';
			q = g_malloc0 (sizeof (ProfileList));
			q->data = data;
			q->prev = p;
			if (p != NULL)
				p->next = q;
			p = q;
			data = NULL;
		  }
	  }
	if (!fileio_close (fio) || p == NULL)
	  {
		while (p != NULL)
		  {
			q = p->prev;
			g_free (p->data);
			g_free (p);
			p = q;
		  }
		return NULL;
	  }
	while (p->prev != NULL)
		p = p->prev;
	r = p;
	/* ꥹȤ */
	while (p != NULL)
	  {
		data = p->data != NULL ? g_strstrip (g_strdup (p->data)) : NULL;
		n = data != NULL ? g_strlen (data) : 0;
		if (n <= 0)
		  {
			p->type = PROFILE_DATA_TYPE_SPACE;
		  }
		else if (data[0] == '#' || data[0] == ';')
		  {
			p->type = PROFILE_DATA_TYPE_COMMENT;
		  }
		else if (data[0] == '[' && data[n - 1] == ']')
		  {
			p->type = PROFILE_DATA_TYPE_SECTION;
			g_free (section);
			section = g_strdup (data + 1);
			section[n - 2] = '\0';
			for (q = p->prev; q != NULL && q->type == PROFILE_DATA_TYPE_SPACE;
																q = q->prev)
			  {
				g_free (q->section);
				q->section = NULL;
			  }
		  }
		else if (g_strchr (data, '=') != NULL)
		  {
			p->type = PROFILE_DATA_TYPE_KEY;
			p->key = g_strdup (data);
			*g_strchr (p->key, '=') = '\0';
			p->value = g_strchr (p->data, '=') + 1;
		  }
		else
		  {
			p->type = PROFILE_DATA_TYPE_UNKNOW;
		  }
		p->section = g_strdup (section);
		g_free (data);
		p = p->next;
	  }
	g_free (section);
	return r;
}


/*	ե򳫤
	file,ե̾
	 RET,ץե														*/
Profile *
profile_open (const gchar *file)
{
	gchar *sys, *tmp;
	gint n;
	Profile *profile;

	profile = g_malloc0 (sizeof (Profile));
	tmp = g_strdup (file != NULL ? file : g_basename (g_get_prgname ()));
	if (!g_path_is_absolute (tmp))
	  {
		n = g_strlen (tmp);
		while (tmp[0] == '.')
			g_memmove (tmp, tmp + 1, n);
		n = g_strlen (tmp);
		if (n >= 5 && g_strcasecmp (tmp + n - 4, ".exe") == 0)
			tmp[n -= 4] = '\0';
		if (n > 0)
		  {
#ifdef SYSCONFDIR
# ifdef G_OS_WIN32
			sys = g_strconcat (SYSCONFDIR, "\\", tmp, ".conf", NULL);
# else /* not G_OS_WIN32 */
			sys = g_strconcat (SYSCONFDIR, "/", tmp, ".conf", NULL);
# endif /* not G_OS_WIN32 */
#else /* not SYSCONFDIR */
			gchar *dir;

			dir = g_dirname (g_get_prgname ());
# ifdef G_OS_WIN32
			sys = g_strconcat (dir, "\\", tmp, ".conf", NULL);
# else /* not G_OS_WIN32 */
			sys = g_strconcat (dir, "/", tmp, ".conf", NULL);
# endif /* not G_OS_WIN32 */
			g_free (dir);
#endif /* not SYSCONFDIR */
			profile->sublist = profile_list (sys);
			g_free (sys);
		  }
	  }
	g_free (tmp);
	tmp = g_strdup (g_basename (g_get_prgname ()));
	n = g_strlen (tmp);
	if (n >= 5 && g_strcasecmp (tmp + n - 4, ".exe") == 0)
		tmp[n - 4] = '\0';
#ifdef G_OS_WIN32
	profile->file = file != NULL ? g_path_is_absolute(file) ? g_strdup (file)
					: g_strconcat (g_get_home_dir (), "\\", file, NULL)
					: g_strconcat (g_get_home_dir (), "\\", tmp, ".ini", NULL);
#else /* not G_OS_WIN32 */
	profile->file = file != NULL ? g_path_is_absolute(file) ? g_strdup (file)
							: g_strconcat (g_get_home_dir (), "/", file, NULL)
							: g_strconcat (g_get_home_dir (), "/.", tmp, NULL);
#endif /* not G_OS_WIN32 */
	g_free (tmp);
	profile->list = profile_list (profile->file);
	return profile;
}


/*	եĤ
	profile,ץե
	    RET,TRUE:ｪλ,FALSE:顼										*/
gboolean profile_close(Profile *profile)
{
	FileIO *fio;
	ProfileList *p,*q;

	if (profile==NULL)
		return TRUE;
	if (profile->edit && (fio = fileio_open (profile->file,
										FILEIO_ACCESS_WRITE, FILEIO_SHARE_READ,
											FILEIO_MODE_CREATE_ALWAYS))!=NULL)
	  {
		/* ѹäȤ */
		for (p = profile->list; p != NULL; p = p->next)
		  {
			if (p->data != NULL)
				fileio_puts (p->data, fio);
			fileio_putc ('\n', fio);
		  }
		fileio_close (fio);
	  }
	g_free(profile->file);
	for (p=profile->list;p!=NULL;p=q) {
		q=p->next;
		g_free(p->data);
		g_free(p->section);
		g_free(p->key);
		g_free(p);
	}
	for (p = profile->sublist; p != NULL; p = q)
	  {
		q = p->next;
		g_free (p->data);
		g_free (p->section);
		g_free (p->key);
		g_free (p);
	  }
	g_free (profile);
	return TRUE;
}


/*	ե뤫ʸ
	profile,ץե
	section,
	    key,
	    RET,ʸ,NULL:顼												*/
gchar *profile_get_string(Profile *profile,
										const gchar *section,const gchar *key)
{
	ProfileList *p;

	if (profile==NULL || section==NULL || key==NULL)
		return FALSE;
	for (p=profile->list;p!=NULL;p=p->next)
		if (p->type == PROFILE_DATA_TYPE_KEY 
										&& g_strcmp (p->section, section) == 0
												&& g_strcmp (p->key, key) == 0)
			return g_strdup (p->value);
	for (p = profile->sublist; p != NULL; p = p->next)
		if (p->type == PROFILE_DATA_TYPE_KEY
										&& g_strcmp (p->section, section) == 0
												&& g_strcmp (p->key, key) == 0)
			return g_strdup (p->value);
	return NULL;
}


/*	ե뤫ͤΥ
	profile,ץե
	section,
	    key,
	   type,
	    RET,Хȿ,0:顼												*/
gint profile_get_size(Profile *profile,
						const gchar *section,const gchar *key,const guint type)
{
	guint8 *array;
	gint n;
	ProfileList *p;

	if (profile==NULL || section==NULL || key==NULL)
		return 0;
	for (p=profile->list;p!=NULL;p=p->next)
		if (p->type == PROFILE_DATA_TYPE_KEY
										&& g_strcmp (p->section, section) == 0
												&& g_strcmp (p->key, key) == 0)
			break;
	if (p == NULL)
		for (p = profile->sublist; p != NULL; p = p->next)
			if (p->type == PROFILE_DATA_TYPE_KEY
										&& g_strcmp (p->section, section) == 0
												&& g_strcmp (p->key, key) == 0)
			break;
	if (p==NULL)
		return 0;
	switch (type) {
		case PROFILE_VALUE_TYPE_BOOL:
			return g_strcmp (p->value, "true") == 0
				|| g_strcmp (p->value, "false") == 0 ? sizeof (gboolean) : 0;
		case PROFILE_VALUE_TYPE_INT:
			return sizeof(gint);
		case PROFILE_VALUE_TYPE_STRING:
			return g_strlen(p->value)+1;
		case PROFILE_VALUE_TYPE_ARRAY:
			if ((array=misc_str_to_array(&n,p->value,10,FALSE))==NULL)
				return 0;
			g_free(array);
			return n;
	}
	return 0;
}


/*	ե뤫ͤ
	profile,ץե
	section,
	    key,
	  value,ͤХåե
	   size,ͤХåեΥ
	   type,
	    RET,TRUE:ｪλ,FALSE:顼										*/
gboolean profile_get_value(Profile *profile,
							const gchar *section,const gchar *key,
							gpointer value,const gint size,const guint type)
{
	guint8 *array;
	gint n;
	ProfileList *p;

	if (profile==NULL || section==NULL || key==NULL || value==NULL)
		return FALSE;
	for (p=profile->list;p!=NULL;p=p->next)
		if (p->type == PROFILE_DATA_TYPE_KEY
										&& g_strcmp (p->section, section) == 0
												&& g_strcmp (p->key, key) == 0)
			break;
	if (p == NULL)
		for (p = profile->sublist; p != NULL; p = p->next)
			if (p->type == PROFILE_DATA_TYPE_KEY
										&& g_strcmp (p->section, section) == 0
												&& g_strcmp (p->key, key) == 0)
			break;
	if (p==NULL)
		return FALSE;
	switch (type) {
		case PROFILE_VALUE_TYPE_BOOL:
			if (size<sizeof(gboolean))
				return FALSE;
			if (g_strcmp (p->value, "true") == 0)
				*((gboolean *)value) = TRUE;
			else if (g_strcmp (p->value, "false") == 0)
				*((gboolean *)value) = FALSE;
			else
				return FALSE;
			break;
		case PROFILE_VALUE_TYPE_INT:
			if (size<sizeof(gint))
				return FALSE;
			misc_str_to_val((gint *)value,p->value,10,TRUE);
			break;
		case PROFILE_VALUE_TYPE_STRING:
			if (size<g_strlen(p->value)+1)
				return FALSE;
			g_strcpy((gchar *)value,p->value);
			break;
		case PROFILE_VALUE_TYPE_ARRAY:
			if ((array=misc_str_to_array(&n,p->value,10,FALSE))==NULL)
				return FALSE;
			if (size<=n)
				g_memmove(value,array,size);
			g_free(array);
			if (n<size)
				return FALSE;
			break;
		default:
			return FALSE;
	}
	return TRUE;
}


/*	եͤꤹ
	profile,ץե
	section,
	    key,
	  value,ͤäƤХåե
	   size,ͤäƤХåեΥ
	   type,
	    RET,TRUE:ｪλ,FALSE:顼										*/
gboolean profile_set_value(Profile *profile,
						const gchar *section,const gchar *key,
						gconstpointer value,const gint size,const guint type)
{
	gchar *data;
	gint i;
	ProfileList *p,*q=NULL;

	if (profile==NULL || section==NULL || key==NULL || value==NULL)
		return FALSE;
	for (p=profile->list;p!=NULL;q=p,p=p->next)
		if (p->type == PROFILE_DATA_TYPE_KEY
										&& g_strcmp (p->section, section) == 0
												&& g_strcmp (p->key, key) == 0)
			break;
	if (p==NULL) {
		for (p = q; p != NULL;p = p->prev)
			if (p->section != NULL && g_strcmp (p->section, section) == 0)
				break;
		if (p==NULL) {
			if (q!=NULL) {
				/* ΥǡȤδ֤˥ڡ */
				p=g_malloc(sizeof(ProfileList));
				p->type=PROFILE_DATA_TYPE_SPACE;
				p->data=p->section=p->key=p->value=NULL;
				p->prev=q;
				p->next=q->next;
				q->next=p;
				q=p;
			}
			/* ⥭ʤȤˤϥ */
			p=g_malloc(sizeof(ProfileList));
			p->type=PROFILE_DATA_TYPE_SECTION;
			p->data=g_strdup_printf("[%s]",section);
			p->section=g_strdup(section);
			p->key=p->value=NULL;
			p->prev=q;
			if (q!=NULL) {
				p->next=q->next;
				q->next=p;
			} else {
				p->next=NULL;
				profile->list=p;
			}
		}
		q=p;
		while (q->type == PROFILE_DATA_TYPE_SPACE && q->section != NULL
					&& g_strcmp (p->section, section) == 0 && q->prev != NULL)
			q = q->prev;
		/* κǸ˥ */
		p=g_malloc(sizeof(ProfileList));
		p->type=PROFILE_DATA_TYPE_KEY;
		p->data=g_strdup_printf("%s=",key);
		p->section=g_strdup(section);
		p->key=g_strdup(key);
		p->value=g_strchr(p->data,'=')+1;
		p->prev=q;
		p->next=q->next;
		q->next=p;
		if (p->next!=NULL)
			p->next->prev=p;
	}
	switch (type) {
		case PROFILE_VALUE_TYPE_BOOL:
			g_free(p->data);
			p->data=g_strdup_printf("%s=%s",
								p->key,*((gboolean *)value)?"true":"false");
			break;
		case PROFILE_VALUE_TYPE_INT:
			g_free(p->data);
			p->data=g_strdup_printf("%s=%d",p->key,*((gint *)value));
			break;
		case PROFILE_VALUE_TYPE_STRING:
			g_free(p->data);
			p->data=g_strdup_printf("%s=%s",p->key,(gchar *)value);
			break;
		case PROFILE_VALUE_TYPE_ARRAY:
			g_free(p->data);
			p->data=g_strdup_printf("%s=%u",p->key,((guint8 *)value)[0]);
			for (i=1;i<size;i++) {
				data=g_strdup_printf("%s %u",p->data,((guint8 *)value)[i]);
				g_free(p->data);
				p->data=data;
			}
			break;
		default:
			return FALSE;
	}
	p->value=g_strchr(p->data,'=')+1;
	profile->edit=TRUE;
	return TRUE;
}


/*	եΥ
	profile,ץե
	section,
	    RET,TRUE:ｪλ,FALSE:顼										*/
gboolean profile_delete_section(Profile *profile,const gchar *section)
{
	gboolean result=FALSE;
	ProfileList *p,*q;

	if (profile==NULL || section==NULL)
		return FALSE;
	for (p=profile->list;p!=NULL;p=q) {
		q=p->next;
		if (p->section != NULL && g_strcmp (p->section, section) == 0)
		  {
			if (p->prev!=NULL)
				p->prev->next=p->next;
			if (p->next!=NULL)
				p->next->prev=p->prev;
			g_free(p->data);
			g_free(p->section);
			g_free(p->key);
			g_free(p);
			profile->edit=TRUE;
			result=TRUE;
		  }
	}
	return result;
}


/*	եΥ
	profile,ץե
	section,
	    key,
	    RET,TRUE:ｪλ,FALSE:顼										*/
gboolean profile_delete_key(Profile *profile,
										const gchar *section,const gchar *key)
{
	gboolean result=FALSE;
	ProfileList *p,*q;

	if (profile==NULL || section==NULL || key==NULL)
		return FALSE;
	for (p=profile->list;p!=NULL;p=q) {
		q=p->next;
		if (p->section != NULL && p->key != NULL
										&& g_strcmp (p->section, section) == 0
												&& g_strcmp (p->key, key) == 0)
		  {
			if (p->prev!=NULL)
				p->prev->next=p->next;
			if (p->next!=NULL)
				p->next->prev=p->prev;
			g_free(p->data);
			g_free(p->section);
			g_free(p->key);
			g_free(p);
			profile->edit=TRUE;
			result=TRUE;
		  }
	}
	return result;
}


/*	եΥ󤹤
	profile,ץե
	    RET,Υꥹ,NULL:顼									*/
GList *
profile_enum_section (Profile *profile)
{
	GList *glist = NULL;
	ProfileList *p;

	if (profile == NULL)
		return FALSE;
	for (p = profile->list; p != NULL; p = p->next)
		if (p->section != NULL
					&& (glist == NULL || g_list_find_custom (glist, p->section,
												(GCompareFunc)strcmp) == NULL))
			glist = g_list_insert_sorted (glist, p->section,
														(GCompareFunc)strcmp);
	for (p = profile->sublist; p != NULL; p = p->next)
		if (p->section != NULL
					&& (glist == NULL || g_list_find_custom (glist, p->section,
												(GCompareFunc)strcmp) == NULL))
			glist = g_list_insert_sorted (glist, p->section,
														(GCompareFunc)strcmp);
	return glist;
}


/*	եΥ󤹤
	profile,ץե
	section,
	    RET,Υꥹ,NULL:顼									*/
GList *
profile_enum_key (Profile *profile, const gchar *section)
{
	GList *glist = NULL;
	ProfileList *p;

	if (profile == NULL)
		return FALSE;
	for (p = profile->list; p != NULL; p = p->next)
		if (p->section != NULL && p->key != NULL
										&& g_strcmp (p->section, section) == 0
			&& (glist == NULL || g_list_find_custom (glist, p->key,
												(GCompareFunc)strcmp) == NULL))
			glist = g_list_insert_sorted (glist, p->key, (GCompareFunc)strcmp);
	for (p = profile->sublist; p != NULL; p = p->next)
		if (p->section != NULL && p->key != NULL
										&& g_strcmp (p->section, section) == 0
			&& (glist == NULL || g_list_find_custom (glist, p->key,
												(GCompareFunc)strcmp) == NULL))
			glist = g_list_insert_sorted (glist, p->key, (GCompareFunc)strcmp);
	return glist;
}
