/*
    Text maid
    copyright (c) 1998-2004 Kazuki IWAMOTO http://www.maid.org/ iwm@maid.org

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "charset.h"
#include "command.h"
#include "edit.h"
#include "find.h"
#include "general.h"
#include "sigfile.h"
#include "toolbar.h"
#include "misc/misc.h"


/******************************************************************************
*                                                                             *
* 一般関数群                                                                  *
*                                                                             *
******************************************************************************/
/*  キャレットを消す
    tmaid,TXTウインドウ情報                                                 */
void hide_caret(TmaidWindow *tmaid)
{
    GdkRectangle rc;

    gtk_timeout_remove(timer_id);
    timer_id=0;
    if (caret) {
        caret=FALSE;
        rc.x=edit_get_align_pos(tmaid,tmaid->cursor.x,tmaid->cursor.y,FALSE);
        rc.y=tmaid->cursor.y;
        rc.x=(rc.x-tmaid->top.x)*tmaid->font_width;
        rc.y=(rc.y-tmaid->top.y)*tmaid->font_height;
        rc.width=tmaid->font_width;
        rc.height=tmaid->font_height;
        gtk_widget_draw(tmaid->drawing,&rc);
    }
}


/*  キャレットの描画
     tmaid,TXTウインドウ情報
    cursor,古いキャレット位置                                               */
void draw_caret(TmaidWindow *tmaid,GdkPoint *cursor)
{
    gboolean visible;
    gchar *text;
    gint x,y;
    guint context_id;
    GdkRectangle rc;
    GtkWidget *widget;

    if (cursor!=NULL && timer_id!=0
            && (tmaid->cursor.x!=cursor->x || tmaid->cursor.y!=cursor->y)) {
        /* タイマ再設定 */
        gtk_timeout_remove(timer_id);
        timer_id=0;
        /* 古いキャレットを消去 */
        rc.x=edit_get_align_pos(tmaid,cursor->x,cursor->y,FALSE);
        rc.y=cursor->y;
        rc.x=(rc.x-tmaid->top.x)*tmaid->font_width;
        rc.y=(rc.y-tmaid->top.y)*tmaid->font_height;
        rc.width=tmaid->font_width;
        rc.height=tmaid->font_height;
        gtk_widget_draw(tmaid->drawing,&rc);
    }
    /* 新しいキャレットを描画 */
    if (timer_id==0)
        caret=TRUE;
    x=edit_get_align_pos(tmaid,tmaid->cursor.x,tmaid->cursor.y,FALSE);
    y=tmaid->cursor.y;
    rc.x=(x-tmaid->top.x)*tmaid->font_width;
    rc.y=(y-tmaid->top.y)*tmaid->font_height;
    rc.width=tmaid->font_width;
    rc.height=tmaid->font_height;
    gtk_widget_draw(tmaid->drawing,&rc);
    if (timer_id==0)
        timer_id=gtk_timeout_add(500,signal_timeout,NULL);
    text = g_strdup_printf (_("L%d,C%d"), y + 1, x + 1);
    gtk_label_set_text (GTK_LABEL (label_cursor), text);
    g_free (text);
    text = tmaid->select.x >= 0
                    ? g_strdup_printf (_("Selected(L%d,C%d)-(L%d,C%d)"),
                        tmaid->select.y + 1, tmaid->select.x + 1, y + 1, x + 1)
                    : g_strdup ("");
    context_id=gtk_statusbar_get_context_id(GTK_STATUSBAR(status),"Status");
    gtk_statusbar_pop(GTK_STATUSBAR(status),context_id);
    gtk_statusbar_push(GTK_STATUSBAR(status),context_id,text);
    g_free(text);
    gdk_window_get_origin (tmaid->drawing->window, &x, &y);
    widget = gtk_widget_get_parent (tmaid->preedit);
    visible = GTK_WIDGET_VISIBLE (widget);
    if (!visible)
      {
        gtk_window_resize (GTK_WINDOW (widget), 1, 1);
        gtk_widget_show (widget);
      }
    gtk_window_move (GTK_WINDOW (widget), rc.x + x, rc.y + y);
    if (!visible)
        gtk_widget_hide (widget);
    rc.width = 0;
    gtk_im_context_set_cursor_location (tmaid->im_context, &rc);
}


/*  メニューを設定する
    tmaid,TXTウインドウ情報                                                 */
void set_menu_bar(TmaidWindow *tmaid)
{
    GList *glist;
    GtkWidget *menu;

    if (tmaid!=NULL) {
        gtk_widget_set_sensitive (gtk_menu_get_attach_widget (GTK_MENU
            (gtk_item_factory_get_widget (ifactory_menu, "<main>/Edit"))),
                                                                        TRUE);
        gtk_widget_set_sensitive (gtk_menu_get_attach_widget (GTK_MENU
            (gtk_item_factory_get_widget (ifactory_menu, "<main>/Search"))),
                                                                        TRUE);
        gtk_widget_set_sensitive (gtk_item_factory_get_widget (ifactory_menu,
                                "<main>/File/Close"), TRUE);
        gtk_widget_set_sensitive (gtk_item_factory_get_widget (ifactory_menu,
                                "<main>/File/Save"), TRUE);
        gtk_widget_set_sensitive (gtk_item_factory_get_widget (ifactory_menu,
                                "<main>/File/Save As..."), TRUE);
        gtk_widget_set_sensitive (gtk_item_factory_get_widget (ifactory_menu,
                                "<main>/File/Reload"), !tmaid->create);
        gtk_widget_set_sensitive (gtk_item_factory_get_widget (ifactory_menu,
                                "<main>/File/Print..."), TRUE);
        gtk_widget_set_sensitive (gtk_item_factory_get_widget (ifactory_menu,
                                "<main>/File/Properties..."), TRUE);
        gtk_widget_set_sensitive (gtk_item_factory_get_widget (ifactory_menu,
                                "<main>/Edit/Undo"), tmaid->undo != NULL);
        gtk_widget_set_sensitive (gtk_item_factory_get_widget (ifactory_menu,
                                "<main>/Edit/Redo"), tmaid->redo != NULL);
        gtk_widget_set_sensitive (gtk_item_factory_get_widget (ifactory_menu,
                                "<main>/Edit/Cut"), tmaid->select.x >= 0);
        gtk_widget_set_sensitive (gtk_item_factory_get_widget (ifactory_menu,
                                "<main>/Edit/Copy"), tmaid->select.x >= 0);
        gtk_widget_set_sensitive (gtk_item_factory_get_widget (ifactory_menu,
                                "<main>/Edit/Delete"), tmaid->select.x >= 0);
        gtk_widget_set_sensitive (gtk_item_factory_get_widget (ifactory_menu,
                                "<main>/Search/Next"), find_num > 0);
        gtk_widget_set_sensitive (gtk_item_factory_get_widget (ifactory_menu,
                                "<main>/Search/Previous"), find_num > 0);
        gtk_widget_set_sensitive(toolbar_items[ 2].item,TRUE);
        gtk_widget_set_sensitive(toolbar_items[ 4].item,TRUE);
        gtk_widget_set_sensitive(toolbar_items[ 6].item,TRUE);
        gtk_widget_set_sensitive(toolbar_items[ 8].item,tmaid->select.x>=0);
        gtk_widget_set_sensitive(toolbar_items[ 9].item,tmaid->select.x>=0);
        gtk_widget_set_sensitive(toolbar_items[10].item,TRUE);
        gtk_widget_set_sensitive(toolbar_items[11].item,tmaid->select.x>=0);
        gtk_widget_set_sensitive(toolbar_items[13].item,tmaid->undo!=NULL);
        gtk_widget_set_sensitive(toolbar_items[14].item,tmaid->redo!=NULL);
        gtk_widget_set_sensitive(toolbar_items[16].item,TRUE);
        gtk_widget_set_sensitive(toolbar_items[17].item,TRUE);
#ifdef G_OS_WIN32
        gtk_widget_set_sensitive (toolbar_items[19].item, TRUE);
#endif /* G_OS_WIN32 */
    } else {
        gtk_widget_set_sensitive (gtk_menu_get_attach_widget (GTK_MENU
            (gtk_item_factory_get_widget (ifactory_menu, "<main>/Edit"))),
                                                                        FALSE);
        gtk_widget_set_sensitive (gtk_menu_get_attach_widget (GTK_MENU
            (gtk_item_factory_get_widget (ifactory_menu, "<main>/Search"))),
                                                                        FALSE);
        gtk_widget_set_sensitive (gtk_item_factory_get_widget (ifactory_menu,
                                "<main>/File/Close"), FALSE);
        gtk_widget_set_sensitive (gtk_item_factory_get_widget (ifactory_menu,
                                "<main>/File/Save"), FALSE);
        gtk_widget_set_sensitive (gtk_item_factory_get_widget (ifactory_menu,
                                "<main>/File/Save As..."), FALSE);
        gtk_widget_set_sensitive (gtk_item_factory_get_widget (ifactory_menu,
                                "<main>/File/Reload"), FALSE);
        gtk_widget_set_sensitive (gtk_item_factory_get_widget (ifactory_menu,
                                "<main>/File/Print..."), FALSE);
        gtk_widget_set_sensitive (gtk_item_factory_get_widget (ifactory_menu,
                                "<main>/File/Properties..."), FALSE);
        gtk_widget_set_sensitive(toolbar_items[ 2].item,FALSE);
        gtk_widget_set_sensitive(toolbar_items[ 4].item,FALSE);
        gtk_widget_set_sensitive(toolbar_items[ 6].item,FALSE);
        gtk_widget_set_sensitive(toolbar_items[ 8].item,FALSE);
        gtk_widget_set_sensitive(toolbar_items[ 9].item,FALSE);
        gtk_widget_set_sensitive(toolbar_items[10].item,FALSE);
        gtk_widget_set_sensitive(toolbar_items[11].item,FALSE);
        gtk_widget_set_sensitive(toolbar_items[13].item,FALSE);
        gtk_widget_set_sensitive(toolbar_items[14].item,FALSE);
        gtk_widget_set_sensitive(toolbar_items[16].item,FALSE);
        gtk_widget_set_sensitive(toolbar_items[17].item,FALSE);
#ifdef G_OS_WIN32
        gtk_widget_set_sensitive (toolbar_items[19].item, FALSE);
#endif /* G_OS_WIN32 */
    }
    menu = gtk_item_factory_get_widget (ifactory_menu, "<main>/Window");
    glist=gtk_container_children(GTK_CONTAINER(menu));
    gtk_widget_set_sensitive(gtk_menu_get_attach_widget(GTK_MENU(menu)),
                                                    g_list_length(glist)>0);
    g_list_free(glist);
}


/*  指定範囲を再描画する
    tmaid,TXTウインドウ情報
    start,開始
      end,終了                                                              */
void clear_sel(TmaidWindow *tmaid,GdkPoint *start,GdkPoint *end)
{
    gint st,ed;
    GdkRectangle rc;

    if (start->y==end->y) {
        /* 同じ行 */
        st=edit_get_align_pos(tmaid,MIN(start->x,end->x),start->y,FALSE);
        ed=edit_get_align_pos(tmaid,MAX(start->x,end->x),start->y,TRUE);
        rc.x=(st-tmaid->top.x)*tmaid->font_width;
        rc.y=(start->y-tmaid->top.y)*tmaid->font_height;
        rc.width=(ed-st+1)*tmaid->font_width;
        rc.height=tmaid->font_height;
    } else {
        /* 違う行 */
        rc.x=0;
        rc.y=(MIN(start->y,end->y)-tmaid->top.y)*tmaid->font_height;
        rc.width=tmaid->drawing->allocation.width;
        rc.height=(ABS(start->y-end->y)+1)*tmaid->font_height;
    }
    gtk_widget_draw(tmaid->drawing,&rc);
}


/*  移動する
    tmaid,TXTウインドウ情報
    top,古い左上の座標                                                      */
void move_edit_window(TmaidWindow *tmaid,GdkPoint *top)
{
    if (tmaid->top.x!=top->x || tmaid->top.y!=top->y)
        misc_scroll_window(tmaid->drawing,
                                    (top->x-tmaid->top.x)*tmaid->font_width,
                                    (top->y-tmaid->top.y)*tmaid->font_height,
                                    0,0,
                                    tmaid->drawing->allocation.width,
                                    tmaid->drawing->allocation.height);
}


/*  キャレット位置にデータを挿入/上書きする
     tmaid,TXTウインドウ情報
      text,データを格納するポインタ
    length,データのバイト数
     caret,TRUE:移動する,FALSE:移動しない
    select,TRUE:選択する,FALSE:選択しない
       RET,逆の操作を行うDOING構造体,NULL:エラー                            */
TmaidHistory *edit_operation(TmaidWindow *tmaid,
                                    const gchar *text,const gint length,
                                    const gboolean caret,const gboolean select)
{
    gint i,j,sx,sy,max,delete,put;
    GdkPoint cursor_old,cursor_new,top;
    GdkRectangle rc;
    LineBuffer *p,*q;
    TmaidHistory *d;

    hide_caret(tmaid);
    tmaid->cursor.x=edit_get_align_pos(tmaid,
                                        tmaid->cursor.x,tmaid->cursor.y,FALSE);
    top=tmaid->top;
    sx=MAX(tmaid->drawing->allocation.width/tmaid->font_width,1);
    sy=MAX(tmaid->drawing->allocation.height/tmaid->font_height,1);
    d=g_malloc(sizeof(TmaidHistory));
    if (tmaid->select.x<0) {
        d->text=NULL;
        d->length=0;
        d->caret=FALSE;
    } else {
        /* 削除 */
        d->length=edit_get_sel_bytes(tmaid,&tmaid->cursor,&tmaid->select);
        d->text=g_malloc(d->length*sizeof(gchar));
        edit_cpy_sel_mem(tmaid,&tmaid->cursor,&tmaid->select,d->text);
        delete=edit_del_sel_mem(tmaid,&tmaid->cursor,&tmaid->select);
        tmaid->max-=delete;/* 行数減少 */
        /* 選択範囲解除とキャレット移動 */
        if (tmaid->select.y<tmaid->cursor.y
                                        || (tmaid->select.y==tmaid->cursor.y
                                        && tmaid->select.x<tmaid->cursor.x)) {
            tmaid->cursor=tmaid->select;
            d->caret=TRUE;
        } else {
            tmaid->cursor=tmaid->cursor;
            d->caret=FALSE;
        }
        tmaid->select.x=-1;
        /* 削除した行の削除した桁より右を初期化(右マージンのために1つ多い) */
        rc.x=(tmaid->cursor.x-tmaid->top.x-1)*tmaid->font_width;
        rc.y=(tmaid->cursor.y-tmaid->top.y)*tmaid->font_height;
        rc.width=tmaid->drawing->allocation.width-rc.x;
        rc.height=tmaid->font_height;
        gtk_widget_draw(tmaid->drawing,&rc);
        /* 減った行数分だけスクロール */
        misc_scroll_window(tmaid->drawing,0,-delete*tmaid->font_height,
                0,
                (tmaid->cursor.y-tmaid->top.y+1)*tmaid->font_height,
                tmaid->drawing->allocation.width,
                MAX(tmaid->drawing->allocation.height
                    -(tmaid->cursor.y-tmaid->top.y+1)*tmaid->font_height,0));
        /* 削除を終えた位置から右マージンの影響のある行数だけ初期化 */
        for (i=0,
                p=edit_get_line_buf(&tmaid->start,&tmaid->off,tmaid->cursor.y);
                                                        p!=NULL;i++,p=p->next)
            if (!p->margin)
                break;
        if (i>0) {
            rc.x=0;
            rc.y=(tmaid->cursor.y-tmaid->top.y+1)*tmaid->font_height;
            rc.width=tmaid->drawing->allocation.width;
            rc.height=i*tmaid->font_height;
            gtk_widget_draw(tmaid->drawing,&rc);
        }
    }
    if (length<=0 || text==NULL) {
        d->cursor=tmaid->cursor;
        d->select.x=-1;
    } else {
        /* 挿入 */
        tmaid->cursor.x=edit_get_align_pos(tmaid,
                                        tmaid->cursor.x,tmaid->cursor.y,FALSE);
        put=edit_put_mem(tmaid,&tmaid->cursor,&cursor_new,text,length);
        tmaid->max+=put;/* 行数増加 */
        cursor_old=tmaid->cursor;
        if (caret) {
            /* キャレット移動 */
            d->cursor=cursor_new;
            d->select=cursor_old;
            tmaid->cursor=cursor_new;
        } else {
            /* キャレット移動なし */
            d->cursor=cursor_old;
            d->select=cursor_new;
        }
        if (select) {
            /* 選択する */
            tmaid->select=d->select;
            gtk_selection_owner_set(tmaid->drawing,GDK_SELECTION_PRIMARY,
                                                            GDK_CURRENT_TIME);
        }
        /* 挿入した行の挿入した桁より右を初期化 */
        rc.x=(cursor_old.x-tmaid->top.x)*tmaid->font_width;
        rc.y=(cursor_old.y-tmaid->top.y)*tmaid->font_height;
        rc.width=tmaid->drawing->allocation.width;
        rc.height=tmaid->font_height;
        gtk_widget_draw(tmaid->drawing,&rc);
        /* 増えた行数分だけスクロール */
        misc_scroll_window(tmaid->drawing,0,put*tmaid->font_height,
                    0,
                    (cursor_old.y-tmaid->top.y+1)*tmaid->font_height,
                    tmaid->drawing->allocation.width,
                    MAX(tmaid->drawing->allocation.height
                        -(cursor_old.y-tmaid->top.y+1)*tmaid->font_height,0));
        /* 挿入を終えた位置から右マージンの影響のある行数だけ初期化 */
        p=q=edit_get_line_buf(&tmaid->start,&tmaid->off,cursor_new.y);
        for (i=cursor_new.y,j=0;
                        p->prev!=NULL && p->prev->margin && cursor_old.y<i;
                                                            i--,j++,p=p->prev);
        while (q->next!=NULL && q->margin)
            j++,q=q->next;
        if (j>0) {
            rc.x=0;
            rc.y=(i-tmaid->top.y+1)*tmaid->font_height;
            rc.width=tmaid->drawing->allocation.width;
            rc.height=j*tmaid->font_height;
            gtk_widget_draw(tmaid->drawing,&rc);
        }
    }
    if (tmaid->cursor.x<tmaid->top.x)
        tmaid->top.x=tmaid->cursor.x;
    else if (tmaid->cursor.x-sx+1>tmaid->top.x)
        tmaid->top.x=tmaid->cursor.x-sx+1;
    if (tmaid->cursor.y<tmaid->top.y)
        tmaid->top.y=tmaid->cursor.y;
    else if (tmaid->cursor.y-sy+1>tmaid->top.y)
        tmaid->top.y=tmaid->cursor.y-sy+1;
    if (tmaid->top.y>tmaid->max-sy)
        tmaid->top.y=MAX(tmaid->max-sy,0);
    max=edit_get_width_max(tmaid);
    if (tmaid->top.x>max-sx+1)
        tmaid->top.x=MAX(max-sx+1,0);
    move_edit_window(tmaid,&top);
    misc_set_scroll_bar (tmaid->hscroll,
                            G_CALLBACK (signal_value_changed_hscroll), tmaid,
                                            0, max + 1, sx, tmaid->top.x);
    misc_set_scroll_bar (tmaid->vscroll,
                            G_CALLBACK (signal_value_changed_vscroll), tmaid,
                                            0, tmaid->max, sy, tmaid->top.y);
    draw_caret(tmaid,NULL);
    return d;
}


/*  リストから削除する
      d,リストの先頭
    RET,削除したリストの数                                                  */
gint delete_list(TmaidHistory **d)
{
    gint count=0;
    TmaidHistory *d0,*d1;

    for (d0=*d;d0!=NULL;d0=d1) {
        g_free(d0->text);
        d1=d0->next;
        g_free(d0);
        count++;
    }
    *d=NULL;
    return count;
}


/*  右マージンで折り返す
    tmaid,TXTウインドウ情報                                                 */
void modify_margin(TmaidWindow *tmaid)
{
    gint data_pos,screen_pos;
    LineBuffer *p,*q;

    /* ラインバッファの先頭へ移動 */
    while (tmaid->start->prev!=NULL)
        tmaid->start=tmaid->start->prev;
    tmaid->off=0;
    p=tmaid->start;
    if (tmaid->ft.limit)
        /* 右マージンによる改行あり */
        while (p!=NULL) {
            data_pos=screen_pos=0;
            while (data_pos<p->length)
                if (p->text[data_pos]=='\t') {
                    screen_pos=(screen_pos/tmaid->ft.tab+1)*tmaid->ft.tab;
                    if (tmaid->ft.margin<screen_pos)
                        break;
                    data_pos++;
                } else if (data_pos+charset_length(p->text[data_pos])
                                                                <=p->length) {
                    screen_pos += charset_width (tmaid->layout, 
                                        p->text + data_pos,
                                        tmaid->font_width, tmaid->font_buf);
                    if (tmaid->ft.margin<screen_pos)
                        break;
                    data_pos+=charset_length(p->text[data_pos]);
                } else {
                    screen_pos++;
                    if (tmaid->ft.margin<screen_pos)
                        break;
                    data_pos++;
                }
            if (tmaid->ft.margin<screen_pos) {
                /* 右マージンを超えているとき、擬似改行する */
                q=g_malloc(sizeof(LineBuffer));
                q->length=p->length-data_pos;
                q->margin=p->margin;
                q->text=g_malloc(q->length*sizeof(gchar));
                q->prev=p;
                q->next=p->next;
                p->next=q;
                if (q->next!=NULL)
                    q->next->prev=q;
                g_memmove(q->text,p->text+data_pos,q->length*sizeof(gchar));
                p->length=data_pos;
                p->margin=TRUE;
                p->text=g_realloc(p->text,p->length*sizeof(gchar));
                p=q;
                tmaid->max++;
            } else if (screen_pos<tmaid->ft.margin
                                            && p->margin && p->next!=NULL) {
                /* 右マージンを下回り擬似改行のとき、次の行とあわせる */
                data_pos=p->length;
                q=p->next;
                if (q->next!=NULL)
                    q->next->prev=p;
                p->next=q->next;
                p->length+=q->length;
                p->margin=q->margin;
                p->text=g_realloc(p->text,p->length*sizeof(gchar));
                g_memmove(p->text+data_pos,q->text,q->length*sizeof(gchar));
                g_free(q->text);
                g_free(q);
                tmaid->max--;
            } else {
                p=p->next;
            }
        }
    else
        /* 右マージンによる改行なし */
        while (p!=NULL)
            if (p->margin) {
                /* 擬似改行のとき、次の行とあわせる */
                data_pos=p->length;
                q=p->next;
                if (q->next!=NULL)
                    q->next->prev=p;
                p->next=q->next;
                p->length+=q->length;
                p->margin=q->margin;
                p->text=g_realloc(p->text,p->length*sizeof(gchar));
                g_memmove(p->text+data_pos,q->text,q->length*sizeof(gchar));
                g_free(q->text);
                g_free(q);
                tmaid->max--;
            } else {
                p=p->next;
            }
}


/*  ja:ファイルの保存を問い合わせる
    tmaid,TXTウインドウ情報
      RET,TRUE:ファイルは閉じる,FALSE:ファイルを閉じない                    */
gboolean
prompt_close (TmaidWindow *tmaid)
{
  if (tmaid->edit)
    {
      gchar *text;

      text = g_strdup_printf (_("File %s was edited.\nSave?"), tmaid->file);
      switch (misc_message_box ("Text maid", text, 0,
                                    _("_Yes"), _("_No"), _("Cancel"), NULL))
        {
          case 0:
            {
              GtkWidget *child;

              child = gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook),
                    gtk_notebook_get_current_page (GTK_NOTEBOOK (notebook)));
              if (g_object_get_data (G_OBJECT (child), "user_data") != tmaid)
                {
                  gint i;
                  GList *glist;

                  glist = gtk_container_children (GTK_CONTAINER (notebook));
                  for (i = g_list_length (glist) - 1; i >= 0; i--)
                    if (g_object_get_data (G_OBJECT (g_list_nth_data
                                            (glist, i)), "user_data") == tmaid)
                      {
                        gtk_notebook_set_current_page (GTK_NOTEBOOK (notebook),
                                                                            i);
                        break;
                      }
                  g_list_free (glist);
                }
              else
                {
                  child = NULL;
                }
              command_save (NULL, 0, NULL);
              if (child)
                gtk_notebook_set_current_page (GTK_NOTEBOOK (notebook),
                    gtk_notebook_page_num (GTK_NOTEBOOK (notebook), child));
              if (!tmaid->edit)
                break;
            }
          case 2:
          case -1:
            g_free (text);
            return FALSE;
        }
      g_free (text);
    }
  return TRUE;
}


/*  タブの編集済みマークの変更
      tmaid,ウインドウ情報
    is_edit,TRUE:編集済み,FALSE:未編集                                      */
void
tmaid_change_edit (TmaidWindow *tmaid, gboolean is_edit)
{
    if (tmaid->edit != is_edit)
      {
        tmaid->edit = is_edit;
        if (is_edit)
            gtk_widget_show (tmaid->mark);
        else
            gtk_widget_hide (tmaid->mark);
      }
}


/******************************************************************************
*                                                                             *
* 高レベル関数群                                                              *
*                                                                             *
******************************************************************************/
/*  編集履歴操作
    tmaid,TXTウインドウ情報
     kind,TRUE:redo,FALSE:undo                                              */
void history_operation(TmaidWindow *tmaid,const gboolean kind)
{
    GdkPoint cursor,select;
    TmaidHistory *d0,*d1;

    if (kind) {
        d0=tmaid->redo;
        tmaid->redo=d0->next;
    } else {
        d0=tmaid->undo;
        tmaid->undo=d0->next;
    }
    cursor=tmaid->cursor;
    select=tmaid->select;
    tmaid->cursor=d0->cursor;
    tmaid->select=d0->select;
    if (select.x>=0)
        clear_sel(tmaid,&select,&cursor);
    d1=edit_operation(tmaid,d0->text,d0->length,d0->caret,TRUE);
    if (kind) {
        d1->next=tmaid->undo;
        tmaid->undo=d1;
    } else {
        d1->next=tmaid->redo;
        tmaid->redo=d1;
    }
    g_free(d0->text);
    g_free(d0);
    set_menu_bar(tmaid);
}


/*  右マージンで改行する
    tmaid,TXTウインドウ情報                                                 */
void margin_operation(TmaidWindow *tmaid)
{
    gboolean limit;
    LineBuffer *p;

    delete_list(&tmaid->undo);
    delete_list(&tmaid->redo);
    limit=tmaid->ft.limit;
    tmaid->ft.limit=TRUE;
    modify_margin(tmaid);
    for (p=tmaid->start;p->prev!=NULL;p=p->prev);
    while (p!=NULL) {
        p->margin=FALSE;
        p=p->next;
    }
    tmaid->ft.limit=limit;
    tmaid_change_edit (tmaid, TRUE);
    tmaid->cursor.x=tmaid->cursor.y=tmaid->top.x=tmaid->top.y=0;
    tmaid->select.x=-1;
    misc_set_scroll_bar (tmaid->hscroll,
            G_CALLBACK (signal_value_changed_hscroll), tmaid,
            0, edit_get_width_max (tmaid) + 1,
            MAX (tmaid->drawing->allocation.width / tmaid->font_width, 1),
                                                                tmaid->top.x);
    misc_set_scroll_bar (tmaid->vscroll,
            G_CALLBACK (signal_value_changed_vscroll), tmaid,
            0, tmaid->max,
            MAX (tmaid->drawing->allocation.height / tmaid->font_height, 1),
                                                                tmaid->top.y);
    set_menu_bar(tmaid);
    gtk_widget_draw(tmaid->drawing,NULL);
}


/*  タブをスペースに変換する
    tmaid,TXTウインドウ情報                                                 */
void tab_operation(TmaidWindow *tmaid)
{
    gint data_pos,screen_pos;
    LineBuffer *p;

    delete_list(&tmaid->undo);
    delete_list(&tmaid->redo);
    for (p=tmaid->start;p->prev!=NULL;p=p->prev);
    while (p!=NULL) {
        data_pos=screen_pos=0;
        while (data_pos<p->length)
            if (p->text[data_pos]=='\t') {
                gint space;

                space = (screen_pos / tmaid->ft.tab + 1) * tmaid->ft.tab
                                                                - screen_pos;
                if (space > 1)
                  {
                    p->text = g_realloc (p->text, (p->length + space - 1)
                                                            * sizeof (gchar));
                    g_memmove (p->text + data_pos + space,
                                p->text + data_pos + 1,
                                (p->length - data_pos - 1) * sizeof (gchar));
                    p->length += space - 1;
                  }
                g_memset (p->text + data_pos, ' ', space * sizeof (gchar));
                screen_pos += space;
                data_pos += space;
            } else if (data_pos+charset_length(p->text[data_pos])<=p->length) {
                screen_pos += charset_width (tmaid->layout, p->text + data_pos,
                                        tmaid->font_width, tmaid->font_buf);
                data_pos+=charset_length(p->text[data_pos]);
            } else {
                screen_pos++;
                data_pos++;
            }
        p=p->next;
    }
    tmaid_change_edit (tmaid, TRUE);
    tmaid->select.x=-1;
    set_menu_bar(tmaid);
    gtk_widget_draw(tmaid->drawing,NULL);
}
