/*
    Text maid
    copyright (c) 1998-2004 Kazuki IWAMOTO http://www.maid.org/ iwm@maid.org

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "argument.h"
#include "charset.h"
#include "edit.h"
#include "file.h"
#include "general.h"
#include "jump.h"
#include "sigfile.h"
#include "sigmain.h"
#include "misc/misc.h"


/******************************************************************************
*                                                                             *
* シグナル/イベント関数群(テキスト)                                           *
*                                                                             *
******************************************************************************/
void signal_style_set(GtkWidget *widget,GtkStyle *style,gpointer user_data)
{
    gint i;
    GList *glist;
    TmaidWindow *tmaid;

    /* システム色 */
    style=gtk_widget_get_style(widget);
    system_color[0]=style->text[0];
    system_color[1]=style->base[0];
    system_color[2]=style->dark[0];
    system_color[3]=style->dark[0];
    system_color[4]=style->dark[0];
    system_color[5]=style->dark[0];
    system_color[6]=style->dark[0];
    system_color[7]=style->mid[0];
    system_color[8]=style->fg[3];
    system_color[9]=style->bg[3];
    if (system_font!=NULL)
        pango_font_description_free (system_font);
    /* システムフォント */
    system_font = pango_font_description_copy
                                    (pango_context_get_font_description
                                    (gtk_widget_get_pango_context (widget)));
    glist=gtk_container_children(GTK_CONTAINER(notebook));
    for (i=g_list_length(glist)-1;i>=0;i--) {
        tmaid = g_object_get_data (G_OBJECT (g_list_nth_data (glist, i)),
                                                                "user_data");
        if (tmaid->ft.syscol || tmaid->ft.font_name==NULL) {
            if (tmaid->ft.font_name==NULL) {
                /* フォント変更 */
                g_object_unref (tmaid->layout);
                gtk_widget_modify_font (tmaid->drawing,
                                    pango_font_description_copy (system_font));
                tmaid->layout = gtk_widget_create_pango_layout
                                                        (tmaid->drawing, NULL);
                charset_size (tmaid->layout,
                    &tmaid->font_width, &tmaid->font_height, tmaid->font_buf);
                /* スクロールバー */
                tmaid->cursor.x=tmaid->cursor.y=tmaid->top.x=tmaid->top.y=0;
                modify_margin(tmaid);
                misc_set_scroll_bar (tmaid->hscroll,
                    G_CALLBACK (signal_value_changed_hscroll), tmaid,
                    0, edit_get_width_max (tmaid) + 1,
                    MAX (tmaid->drawing->allocation.width / tmaid->font_width,
                                                            1), tmaid->top.x);
                misc_set_scroll_bar (tmaid->vscroll,
                    G_CALLBACK (signal_value_changed_vscroll), tmaid,
                    0, tmaid->max,
                    MAX (tmaid->drawing->allocation.height
                                    / tmaid->font_height, 1), tmaid->top.y);
            }
            gtk_widget_draw(tmaid->drawing,NULL);/* 再描画 */
        }
    }
    g_list_free(glist);
}


/******************************************************************************
*                                                                             *
* シグナル/イベント関数群(メニュー)                                           *
*                                                                             *
******************************************************************************/
void signal_activate_menu_window(GtkWidget *widget,GtkWidget *child)
{
    gtk_notebook_set_page(GTK_NOTEBOOK(notebook),
                        gtk_notebook_page_num(GTK_NOTEBOOK(notebook),child));
    gtk_widget_grab_focus (((TmaidWindow *)
                g_object_get_data (G_OBJECT (child), "user_data"))->drawing);
}


void signal_activate_menu_history(GtkWidget *widget,gchar *file)
{
    file_open_edit(file);
}


void signal_destroy_menu_history(GtkWidget *widget,gchar *file)
{
    g_free(file);
}


/******************************************************************************
*                                                                             *
* シグナル/イベント関数群(ノート)                                             *
*                                                                             *
******************************************************************************/
void signal_clicked(GtkWidget *widget,GtkWidget *child)
{
    if (prompt_close ((TmaidWindow *)g_object_get_data (G_OBJECT (child),
                                                                "user_data")))
        gtk_notebook_remove_page (GTK_NOTEBOOK (notebook),
                        gtk_notebook_page_num (GTK_NOTEBOOK (notebook), child));
}


void signal_switch_page(GtkNotebook *notebook,GtkNotebookPage *page,
                                            gint page_num,gpointer user_data)
{
    TmaidWindow *tmaid;

    tmaid = g_object_get_data
                    (G_OBJECT (gtk_notebook_get_nth_page (notebook, page_num)),
                                                                "user_data");
    /* メニュー表示 */
    set_menu_bar(tmaid);
}


void signal_destroy_notebook(GtkWidget *widget,gpointer user_data)
{
    gtk_timeout_remove(timer_id);
}


/******************************************************************************
*                                                                             *
* シグナル/イベント関数群(セレクション)                                       *
*                                                                             *
******************************************************************************/
void
signal_drag_data_received (GtkWidget *widget, GdkDragContext *context,
    gint x, gint y, GtkSelectionData *selection_data, guint info, guint time)
{
    gchar *file, **files;
    gint i;

    files = g_strsplit (selection_data->data, "\r\n", G_MAXINT);
    for (i = 0; files[i] != NULL; i++)
      {
        file = g_filename_from_uri (files[i], NULL, NULL);
        if (file != NULL)
          {
            file_open_edit (file);
            g_free (file);
          }
      }
    g_strfreev (files);
}


gboolean signal_selection_clear(GtkWidget *widget,GdkEventSelection *event,
                                                            gpointer user_data)
{
    if (event->selection == GDK_SELECTION_CLIPBOARD)
      {
        /* クリップボード */
        g_free (clipboard_text);
        clipboard_text = NULL;
      }
    return TRUE;
}


void signal_selection_get(GtkWidget *widget,GtkSelectionData *data,
                                    guint info,guint time,TmaidWindow *tmaid)
{
    gchar *text=NULL;
    gint length=0;

    if (data->selection==GDK_SELECTION_PRIMARY) {
        /* プライマリセレクション */
        if (tmaid!=NULL && tmaid->select.x>=0) {
            length=edit_get_sel_bytes(tmaid,&tmaid->select,&tmaid->cursor);
            text=g_malloc((length+1)*sizeof(gchar));
            edit_cpy_sel_mem(tmaid,&tmaid->select,&tmaid->cursor,text);
            text[length]='\0';
        }
    } else if (data->selection == GDK_SELECTION_CLIPBOARD) {
        /* クリップボード */
        if (clipboard_text!=NULL) {
            text=clipboard_text;
            length=g_strlen(clipboard_text);
        }
    } else if (data->selection==atom_textmaid0) {
        /* プロセス間通信 */
        text="Text maid Selection";
        length=g_strlen(text);
        gtk_selection_convert(window,atom_textmaid1,
                                GDK_SELECTION_TYPE_STRING,GDK_CURRENT_TIME);
    } else if (data->selection==atom_textmaid1 && arg_text!=NULL) {
        /* プロセス間通信 */
        text=arg_text;
        arg_text=NULL;
        length=g_strlen(text);
    }
    if (text!=NULL) {
        gtk_selection_data_set_text (data, text, length);
        if (data->selection==GDK_SELECTION_PRIMARY
                                            || data->selection==atom_textmaid1)
            g_free(text);
    }
}


void signal_selection_received(GtkWidget *widget,GtkSelectionData *data,
                                                guint time,gpointer user_data)
{
    gboolean *arg_files;
    gchar *text = NULL, **argv;
    gint i,count;
    TmaidHistory *d;
    TmaidWindow *tmaid;
    GdkAtom *atoms;
    GdkPoint select;

    if (data->selection == GDK_SELECTION_PRIMARY
                            || data->selection == GDK_SELECTION_CLIPBOARD) {
        /* プライマリセレクション/クリップボード */
        if (data->length>=0) {
            if (data->type==GDK_SELECTION_TYPE_ATOM) {
                atoms=(GdkAtom *)data->data;
                count=data->length/sizeof(GdkAtom);
                for (i = 0; i < count; i++)
                    if (atoms[i] == atom_utf8)
                        break;
                if (i >= count)
                    for (i=0;i<count;i++)
                        if (atoms[i]==atom_compound || atoms[i]==atom_text)
                            break;
                if (i>=count)
                    for (i=0;i<count;i++)
                        if (atoms[i]==GDK_TARGET_STRING)
                            break;
                if (i<count)
                    gtk_selection_convert(widget,data->selection,atoms[i],
                                                            GDK_CURRENT_TIME);
            } else {
                text = gtk_selection_data_get_text (data);
            }
        }
        if (text!=NULL) {
            tmaid = g_object_get_data(G_OBJECT
                    (gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook),
                    gtk_notebook_get_current_page (GTK_NOTEBOOK (notebook)))),
                                                                "user_data");
            select=tmaid->select;
            d = edit_operation (tmaid, text, g_strlen (text),
                            data->selection == GDK_SELECTION_CLIPBOARD, FALSE);
            d->next=tmaid->undo;
            tmaid->undo=d;
            if (delete_list(&tmaid->redo)>0 || d->next==NULL || select.x>=0)
                set_menu_bar(tmaid);
            tmaid_change_edit (tmaid, TRUE);
            g_free(text);
        }
    } else if (data->selection==atom_textmaid0) {
        /* プロセス間通信 */
        instance=MAX(data->length,0);
    } else if (data->selection==atom_textmaid1 && data->length>0) {
        /* プロセス間通信 */
        /* コマンドラインの解析 */
        argv=g_strsplit(data->data,"\n",G_MAXINT);
        init_line=0;
        arg_files=arg_analyse(argv,FALSE);
        for (i = 1; argv[i] != NULL; i++)
            if (arg_files[i]) {
                tmaid=file_open_edit(argv[i]);
                if (init_line!=0)   /* 指定行に移動 */
                    jump_operation(tmaid,init_line);
            }
        g_free(arg_files);
        g_strfreev(argv);
    }
}


/******************************************************************************
*                                                                             *
* シグナル/イベント関数群(メイン)                                             *
*                                                                             *
******************************************************************************/
gboolean signal_delete(GtkWidget *widget,GdkEvent *event,gpointer user_data)
{
    gint page;
    TmaidWindow *tmaid;

    while ((page=gtk_notebook_get_current_page(GTK_NOTEBOOK(notebook)))>=0) {
        tmaid = g_object_get_data (G_OBJECT
                (gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook), page)),
                                                                "user_data");
        if (!prompt_close(tmaid))
            return TRUE;
        gtk_notebook_remove_page(GTK_NOTEBOOK(notebook),page);
    }
    return FALSE;
}


void signal_destroy_window(GtkWidget *widget,gpointer user_data)
{
    gint i;
    GList *glist;
    GtkWidget *menu_shell;

    memset(history_files,0,sizeof(gchar *)*10);
    menu_shell = gtk_item_factory_get_widget (ifactory_menu, "<main>/File");
    glist = gtk_container_children (GTK_CONTAINER (menu_shell));
    for (i=0;i<(gint)g_list_length(glist)-MENUFILE-1;i++)
        history_files[i] = g_strdup (g_object_get_data
                        (G_OBJECT (g_list_nth_data (glist, i + MENUFILE - 1)),
                                                                "user_data"));
    g_list_free(glist);
    gtk_main_quit();
}
